import numpy as np
import matplotlib.pyplot as plt
from profit.sur.backend.gp_functions_old import invert, nll, predict_f, \
    get_marginal_variance_BBQ, wld_get_marginal_variance
from profit.sur.backend.kernels import kern_sqexp
from profit.util.halton import halton

def f(x): return x*np.cos(10*x)

# Custom function to build GP matrix
def build_K(xa, xb, hyp, K):
    for i in np.arange(len(xa)):
        for j in np.arange(len(xb)):
            K[i, j] = kern_sqexp(xa[i], xb[j], hyp[0])

noise_train = 0.01

#ntrain = 20
for ntrain in range(1, 31):
    xtrain = halton(ntrain, 1)
    ftrain = f(xtrain)
    np.random.seed(0)
    ytrain = ftrain + noise_train*np.random.randn(ntrain, 1)

    # GP regression with fixed kernel hyperparameters
    hyp = [0.1, 1e-4]  # l and sig_noise**2

    K = np.empty((ntrain, ntrain))   # train-train
    build_K(xtrain, xtrain, hyp, K)  # writes inside K
    Ky = K + hyp[-1]*np.eye(ntrain)
    Kyinv = invert(Ky, 4, 1e-6)       # using gp_functions.invert

    ntest = 300
    xtest = np.linspace(0, 1, ntest)
    ftest = f(xtest)

    Ks = np.empty((ntrain, ntest))  # train-test
    Kss = np.empty((ntest, ntest))  # test-test
    build_K(xtrain, xtest, hyp, Ks)
    build_K(xtest, xtest, hyp, Kss)

    fmean = Ks.T.dot(Kyinv.dot(ytrain)) # predictive mean

    Ef, varf = predict_f(hyp, xtrain.reshape(-1, 1),
                          ytrain.reshape(-1, 1), xtest.reshape(-1, 1), neig=8)# posterior
    # Estimation and variance
    varf = np.diag(varf)

    # we keep only the diag because the variance is on it, the other terms are covariance

    plt.figure()
    plt.plot(xtrain, ytrain, 'kx')
    plt.plot(xtest, ftest, 'm-')
    plt.plot(xtest, fmean, 'r--')
    axes = plt.gca()
    axes.set_ylim([-1.5, 1])
    plt.title('Random Gaussian Process with '+ str(ntrain) + ' observation(s) hyp = [0.1, 1e-4]')




    plt.fill_between(xtest, # x
                     (fmean.flatten() + 2 * np.sqrt(varf)), # y1
                     (fmean.flatten() - 2 * np.sqrt(varf)), facecolor='blue', alpha=0.4) # y2

    plt.legend(('training', 'reference', 'prediction', 'Posterior Variance'))
    plt.savefig('Random_' + str(ntrain))

    #plt.show()



import numpy as np
import matplotlib.pyplot as plt
import GPy

#This functions generate data corresponding to two outputs
f_output1 = lambda x: 4. * np.cos(x/5.) - .4*x - 35. + np.random.rand(x.size)[:,None] * 2.
f_output2 = lambda x: 6. * np.cos(x/5.) + .2*x + 35. + np.random.rand(x.size)[:,None] * 8.


#{X,Y} training set for each output
X1 = np.random.rand(100)[:,None]; X1=X1*75
X2 = np.random.rand(100)[:,None]; X2=X2*70 + 30
Y1 = f_output1(X1)
Y2 = f_output2(X2)
#{X,Y} test set for each output
Xt1 = np.random.rand(100)[:,None]*100
Xt2 = np.random.rand(100)[:,None]*100
Yt1 = f_output1(Xt1)
Yt2 = f_output2(Xt2)

K1 = GPy.kern.Bias(1)
K2 = GPy.kern.Linear(1)
K3 = GPy.kern.Matern32(1)
lcm = GPy.util.multioutput.LCM(input_dim=1,num_outputs=2,kernels_list=[K1,K2,K3])

m = GPy.models.GPCoregionalizedRegression([X1,X2],[Y1,Y2],kernel=lcm)
m['.*ICM.*var'].unconstrain()
m['.*ICM0.*var'].constrain_fixed(1.)
m['.*ICM0.*W'].constrain_fixed(0)
m['.*ICM1.*var'].constrain_fixed(1.)
m['.*ICM1.*W'].constrain_fixed(0)
m.optimize()

newX = np.linspace(0, 100, 100)[:,None]
newX = np.hstack([newX,np.ones_like(newX)])
noise_dict = {'output_index':newX[:,1:].astype(int)}
ytest = m.predict(newX,Y_metadata=noise_dict)

plt.figure()
plt.plot(newX[:,0],ytest[0])
plt.plot(newX[:,0],ytest[1])
plt.plot(X1, Y1, 'x')



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Wed Jun  22 14:50:32 2020

@author: ert
"""
#%%
import sympy
from sympy import symbols, sqrt, exp, diff, IndexedBase, Idx, Eq
from sympy.utilities.codegen import codegen
import shutil

template = """
subroutine kern_{name}(nd, nx, xa, xb, out)
    implicit none
    integer, intent(in) :: nd, nx
    real(8), intent(in) :: xa(nd), xb(nd, nx)
    real(8), intent(out) :: out(nx)
    integer :: kx, kd
    do kx = 1, nx
        out(kx) = {expr}
    end do
end subroutine kern_{name}

"""

#%% Kernel functions (kernel, 1st and second derivatives)
x = symbols(r'x', real=True)
nd, nx = symbols(r'nd, nx', integer=True)
kd = Idx(r'kd', nd)
kx = Idx(r'kx', nx)
xa = IndexedBase(r'xa', shape=(nd))
xb = IndexedBase(r'xb', shape=(nd, nx))

kern = exp(-x**2/2)
dkern = diff(kern, x).simplify()
d2kern = diff(dkern, x).simplify()



# expr = sympy.fcode(kern).replace('x', 'x')

# code = template.format(name='sqexp', expr=expr.strip())

# with open('kernels_base.f90', 'w') as fout:
#     fout.write(code)



from numpy import *
from matplotlib.pyplot import *
from mpl_toolkits.mplot3d import Axes3D
from time import time

import profit

def f(u): return (u[0]**2-u[0])*cos(2*u[1])   # original model

def traintime(lntrain):
    n = int(10**(lntrain)+1)
    print(n)

    utrain = profit.quasirand(npoint=n, ndim=2)   # quasi-random space-filling training points
    ytrain = f(utrain)                            # output at training points
    t = time()
    fresp = profit.fit(utrain, ytrain)            # response model
    return time() - t

lntrain = 2.5 + 1.0*profit.quasirand(npoint=15, ndim=1).flatten()
lttrain = array([log10(traintime(lnt)) for lnt in lntrain])
#%%
ltresp = profit.fit(lntrain, lttrain)

#%%
figure()
lnpl = linspace(2, 4.5, 100)
ltpl, ltvar = ltresp(lnpl)

plot(lnpl, ltpl)
plot(lntrain, lttrain, 'xk')
fill_between(lnpl.flatten(), 
             array(ltpl).flatten() - 1.96*sqrt(array(ltvar).flatten()), 
             array(ltpl).flatten() + 1.96*sqrt(array(ltvar).flatten()),
             alpha = 0.3)
plot([2,4.5], [0,0], color='tab:green')
plot([2,4.5], [log10(60*60), log10(60*60)], color='tab:red')
xlabel('log10(number of points N)')
ylabel('log10(runtime in seconds)')
grid(True)

# %%



import numpy as np
import matplotlib.pyplot as plt
import seaborn as sns

sns.set()
import time
from profit.sur.backend.gp_functions import predict_f
from profit.sur.backend.python_kernels import RBF

# Fixing the dimensions of the figure
plt.rcParams['figure.figsize'] = [12, 7]
# Definition of the dimension
d = 1
# Number of samples (training set)
n = 50
# Training input (x)
# should be defined as a matrix with 1 columns
x = np.linspace(start=0, stop=1, num=n).reshape((n, 1))


# Definition of the function
def f(x):
    return np.sin(2 * np.pi * x) + np.sin(4 * np.pi * x)


# Plot of the function f
plt.plot(x, f(x))
plt.title('Plot of f(x) = $\sin(2 \pi x) + \sin(4 \pi x)$')
plt.xlabel('x')
plt.ylabel('f(x)')
# Definition of the training data (f_x = observations f(x))
f_x = f(x)
# the noise sigma_n
sigma_n = 1e-2
# Definition of the errors -> we assume that the noise on observations
# follows an independent , identically distributed Gaussian distribution with zero mean and variance sigma_n ^2
epsilon = sigma_n * np.random.randn(n).reshape((-1, 1))
# Observed target variable (f(x) + epsilon)
y = f_x + epsilon
# Plot of the error distribution
#sns.distplot(epsilon)
plt.title('Distribution of $\epsilon$ and the interpolated density')
plt.xlabel('$\epsilon$')
plt.ylabel('Number of values')
# define number of test data points
n_star = 30
# define input test data x_star
x_star = np.linspace(0, 1, n_star).reshape((n_star, 1))
# set the hyper - parameters l and sigma_f
l = 0.2
sigma_f = 1

# Definition of the squared exponential kernel
t = time.time()
kernel = RBF(x, y, l)
elapsed = time.time() - t
print(elapsed)
# Definition of the covariance matrices
t = time.time()
a = np.array([l, sigma_f, sigma_n])
K = RBF(x, x, *a)
K_star2 = RBF(x_star, x_star, *a)
K_star = RBF(x_star, x, *a)
elapsed = time.time() - t
print(elapsed)
# Plot of K
plt.imshow(K)
plt.colorbar()
# Plot K_star2
plt.imshow(K_star2)
plt.colorbar()
# Plot Kstar
plt.imshow(K_star)
plt.colorbar()
# Compute C
C = np.block([[K + (sigma_n / sigma_f) ** 2 * np.eye(n), np.transpose(K_star)], [K_star, K_star2]])
# Plot C
plt.imshow(C)
plt.colorbar()
# Plot function f(x)
plt.figure()
plt.plot(x, f_x, 'r')

# plot of n_prior samples from prior distribution (100 samples). The mean is zero and the covariance is given by K_star2
n_prior = 100
for i in range(0, n_prior):
    f_star = np.random.multivariate_normal(np.zeros(n_star), K_star2)
plt.plot(x_star, f_star, 'b', linewidth=0.5)
# Compute posterior mean and covariance.
t = time.time()
f_bar_star, cov_f_star = predict_f(a, x, y, x_star, RBF, return_full_cov=True, neig=8)
elapsed = time.time() - t
print(elapsed)
# plot the covariance matrix
plt.imshow(cov_f_star)
# Plot of n_prior samples from the posterior distribution and of the true function
for i in range(n_prior):
    f_posterior = np.random.multivariate_normal(f_bar_star[:, 0], cov_f_star)
plt.figure()
plt.plot(x_star, f_posterior)
plt.fill_between(x_star.flatten(),
                 (f_bar_star.flatten() + 2 * np.sqrt(np.diag(cov_f_star))),
                 (f_bar_star.flatten() - 2 * np.sqrt(np.diag(cov_f_star))))



from abc import ABC
from .sur import Surrogate
import numpy as np


class ANN(Surrogate, ABC):
    _defaults = {}

    def __init__(self, config):
        super().__init__()

    def train(self, X, y):
        pass

    def predict(self, Xpred):
        pass

    def save_model(self, path):
        pass

    @classmethod
    def load_model(cls, path):
        pass

    @classmethod
    def from_config(cls, config, base_config):
        pass

    @classmethod
    def handle_subconfig(cls, config, base_config):
        pass


@Surrogate.register('ANN')
class ANNSurrogate(Surrogate):
    from tensorflow.keras.models import Sequential
    from tensorflow.keras.layers import Dense
    """ TODO: Port to PyTorch """
    def __init__(self):
        super().__init__()

    def train(self, x, y):
        """Fits a artificial neural network to input points x and
           model outputs y with scale sigma_f and noise sigma_n"""

        self.model = Sequential()
        self.model.add(Dense(64, input_dim=1, activation='relu'))
        self.model.add(Dense(64, activation='relu'))
        self.model.add(Dense(1, activation='linear'))

        self.model.compile(optimizer='adam', loss='mse', metrics=['mse'])
        self.model.fit(x, y, epochs=128, batch_size=16)
        self.trained = True


    def add_training_data(self, x, y, sigma=None):
        """Adds input points x and model outputs y with std. deviation sigma
           and updates the inverted covariance matrix for the GP via the
           Sherman-Morrison-Woodbury formula"""

        raise NotImplementedError()

    def predict(self, x):
        if not self.trained:
            raise RuntimeError('Need to train() before predict()')

        fpred = []
        fpred.append(self.model.predict(x))
        fpred.append(np.zeros_like(fpred[0]))  # no uncertainty information from ANN
        return fpred


# New classes should be based on PyTorch
import torch.nn as nn
import torch.nn.functional as F
import torch.optim as optim


class Autoencoder(Surrogate, nn.Module):
    """Nonlinear autoencoder with activation functions"""
    def __init__(self, D, d):
        super().__init__()
        self.enc1 = nn.Linear(in_features=D, out_features=D//2)
        self.enc2 = nn.Linear(in_features=D//2, out_features=d)
        self.dec1 = nn.Linear(in_features=d, out_features=D//2)
        self.dec2 = nn.Linear(in_features=D//2, out_features=D)

    def forward(self, x):
        x = F.tanh(self.enc1(x))
        x = F.tanh(self.enc2(x))
        x = F.tanh(self.dec1(x))
        x = F.tanh(self.dec2(x))
        return x

    def train(self, x, learning_rate=0.01, nstep=1000):
        optimizer = optim.Adam(ae.parameters(), learning_rate=0.01)
        for k in range(nstep):
            optimizer.zero_grad()
            output = ae(input)
            loss = criterion(output, target)
            loss.backward()
            optimizer.step()
            print(loss.data)



"""
Created on Wed Aug 19 03:47:54 2020

@author: manal khallaayoune
"""

import numpy as np
import GPy
from GPy.kern.src.stationary import Stationary


class ExpSin(Stationary):
    """
    Exponential of Sinus kernel: 
        Product of 1D Exponential of Sinus kernels

    .. math::

        &k(x,x')_i = \sigma^2 \prod_{j=1}^{dimension} \exp \\bigg( - \\frac{ \sin ( x_{i,j}-x_{i,j}' ) ^2}{2 \ell_j^2} \\bigg)
        
        &x,x' \in \mathcal{M}_{n,dimension}
        
        &k \in \mathcal{M}_{n,n}

    """
    def __init__(self, input_dim, variance=1., lengthscale=None, ARD=False, active_dims=None, name='ExpSin'):
        super(ExpSin, self).__init__(input_dim, variance, lengthscale, ARD, active_dims, name)

    def K_of_r(self, dist):
        n = dist.shape[2]
        s = 0
        for k in range(n):
            s+= np.sin(dist[:,:,k])**2
        return self.variance * np.exp(-s/(2*self.lengthscale**2))
    
    def K(self, X, X2):
        dist = X[:,None,:]-X2[None,:,:]
        return self.K_of_r(dist)

    def dK_dr(self,dist,dimX):
        K = self.K_of_r(dist)
        d1 = dist[:,:,dimX]
        return -K*np.sin(d1)*np.cos(d1)/self.lengthscale**2
    
    def dK_dX(self, X, X2, dimX):
        dist = X[:,None,:]-X2[None,:,:]
        dK_dr = self.dK_dr(dist,dimX)
        return dK_dr
    
    def dK_dX2(self,X,X2,dimX2):
        return -self.dK_dX(X,X2, dimX2)
    
    def dK2_dXdX2(self, X, X2, dimX, dimX2):
        dist = X[:,None,:]-X2[None,:,:]
        K = self.K_of_r(dist)
        dK_dX = self.dK_dX(X,X2,dimX)
        lengthscale2inv = np.ones((X.shape[1]))/(self.lengthscale**2)
        l1 = lengthscale2inv[dimX]
        l2 = lengthscale2inv[dimX2]
        d1 = dist[:,:,dimX]
        d2 = dist[:,:,dimX2]
        s1 = np.sin(d1)
        c1 = np.cos(d1)
        return (dimX!=dimX2)*dK_dX*l2*np.sin(d2)*np.cos(d2) + (dimX==dimX2)*l1*(dK_dX*s1*c1+K*c1**2-K*s1**2)
    



import numpy as np
from numpy import exp, sin
import matplotlib.pyplot as plt
import scipy.integrate as spint
from scipy.optimize import fsolve, fmin, minimize, root, newton
from scipy.linalg import solve_triangular
from scipy.sparse.linalg import eigsh
import scipy
from scipy.integrate import solve_ivp
from param import nm, U0, dtsymp
from kernels import *


def f_kern(x, y, x0, y0, l):
    return kern_num(x,y,x0,y0,l[0], l[1])

def dkdx(x, y, x0, y0, l):
    return dkdx_num(x,y,x0,y0,l[0], l[1])

def dkdy(x, y, x0, y0, l):
    return dkdy_num(x,y,x0,y0,l[0], l[1])

def dkdx0(x, y, x0, y0, l):
    return dkdx0_num(x,y,x0,y0,l[0], l[1])

def dkdy0(x, y, x0, y0, l):
    return dkdy0_num(x,y,x0,y0,l[0], l[1])

def d2kdxdx0(x, y, x0, y0, l):
    return d2kdxdx0_num(x,y,x0,y0,l[0], l[1])

def d2kdydy0(x, y, x0, y0, l):
    return d2kdydy0_num(x,y,x0,y0,l[0], l[1])

def d2kdxdy0(x, y, x0, y0, l):
    return d2kdxdy0_num(x,y,x0,y0,l[0], l[1])

def d2kdydx0(x, y, x0, y0, l):
    return d2kdxdy0(x, y, x0, y0, l)

def d3kdxdx0dy0(x, y, x0, y0, l):
    return d3kdxdx0dy0_num(x,y,x0,y0,l[0], l[1])

def d3kdydy0dy0(x, y, x0, y0, l):
    return d3kdydy0dy0_num(x,y,x0,y0,l[0], l[1])

def d3kdxdy0dy0(x, y, x0, y0, l):
    return d3kdxdy0dy0_num(x,y,x0,y0,l[0], l[1])

def d3kdydx0dy0(x, y, x0, y0, l):
    return d3kdxdy0dy0(x, y, x0, y0, l)

def dkdlx(x, y, x0, y0, l):
    return dkdlx_num(x, y, x0, y0, l[0], l[1])

def dkdly(x, y, x0, y0, l):
    return dkdly_num(x, y, x0, y0, l[0], l[1])

def d3kdxdx0dlx(x, y, x0, y0, l):
    return d3kdxdx0dlx_num(x,y,x0,y0,l[0], l[1])

def d3kdydy0dlx(x, y, x0, y0, l):
    return d3kdydy0dlx_num(x,y,x0,y0,l[0], l[1])

def d3kdxdy0dlx(x, y, x0, y0, l):
    return d3kdxdy0dlx_num(x,y,x0,y0,l[0], l[1])

def d3kdydx0dlx(x, y, x0, y0, l):
    return d3kdxdy0dlx(x, y, x0, y0, l)

def d3kdxdx0dly(x, y, x0, y0, l):
    return d3kdxdx0dly_num(x,y,x0,y0,l[0], l[1])

def d3kdydy0dly(x, y, x0, y0, l):
    return d3kdydy0dly_num(x,y,x0,y0,l[0], l[1])

def d3kdxdy0dly(x, y, x0, y0, l):
    return d3kdxdy0dly_num(x,y,x0,y0,l[0], l[1])

def d3kdydx0dly(x, y, x0, y0, l):
    return d3kdxdy0dly(x, y, x0, y0, l)

def intode(y, t, dtsymp):
    ys = np.zeros([2, len(t)])
    ys[0,0] = y[0]
    ys[1,0] = y[1]

    for kt in range(1,len(t)):
            ys[1, kt]=ys[1, kt-1] - dtsymp*(np.sin(ys[0, kt-1] + np.pi))#-0.1*np.sin(ys[0,kt-1]-0.5*t[kt]))
            ys[0, kt]=ys[0, kt-1] + dtsymp*ys[1, kt]

    # ys[0,:] = np.mod(ys[0,:], 2*np.pi)
    return ys.T

def build_K(xin, x0in, l, K):
    # set up covariance matrix
    N = K.shape[1]//2
    N0 = K.shape[0]//2
    x0 = x0in[0:N]
    x = xin[0:N]
    y0 = x0in[N:2*N]
    y = xin[N:2*N]
    for k in range(N0):
        for lk in range(N):
            K[k,lk] = d2kdxdx0(
                x0[k], y0[k], x[lk], y[lk], l)
            K[N0+k,lk] = d2kdxdy0(
                 x0[k], y0[k], x[lk], y[lk], l)
            K[k,N+lk] = d2kdydx0(
                 x0[k], y0[k], x[lk], y[lk], l)
            K[N0+k,N+lk] = d2kdydy0(
                x0[k], y0[k], x[lk], y[lk], l)

def buildK(x, y, x0, y0, l, K):
    # set up covariance matrix
    N = K.shape[1]//2
    N0 = K.shape[0]//2

    for k in range(N0):
        for lk in range(N):
            K[k,lk] = d2kdxdx0(
                x0[k], y0[k], x[lk], y[lk], l)
            K[N0+k,lk] = d2kdxdy0(
                 x0[k], y0[k], x[lk], y[lk], l)
            K[k,N+lk] = d2kdydx0(
                 x0[k], y0[k], x[lk], y[lk], l)
            K[N0+k,N+lk] = d2kdydy0(
                x0[k], y0[k], x[lk], y[lk], l)


def build_dK(x, y, x0, y0, l):
    # set up covariance matrix
    N = len(x)
    N0 = len(x0)
    k11 = np.empty((N0, N))
    k12 = np.empty((N0, N))
    k21 = np.empty((N0, N))
    k22 = np.empty((N0, N))

    dK = []

    for k in range(N0):
        for lk in range(N):
              k11[k,lk] = d3kdxdx0dlx(
                  x0[k], y0[k], x[lk], y[lk], l)
              k21[k,lk] = d3kdxdy0dlx(
                  x0[k], y0[k], x[lk], y[lk], l)
              k12[k,lk] = d3kdydx0dlx(
                  x0[k], y0[k], x[lk], y[lk], l)
              k22[k,lk] = d3kdydy0dlx(
                  x0[k], y0[k], x[lk], y[lk], l)

    dK.append(np.vstack([
        np.hstack([k11, k12]),
        np.hstack([k21, k22])
    ]))

    for k in range(N0):
        for lk in range(N):
             k11[k,lk] = d3kdxdx0dly(
                 x0[k], y0[k], x[lk], y[lk], l)
             k21[k,lk] = d3kdxdy0dly(
                 x0[k], y0[k], x[lk], y[lk], l)
             k12[k,lk] = d3kdydx0dly(
                  x0[k], y0[k], x[lk], y[lk], l)
             k22[k,lk] = d3kdydy0dly(
                 x0[k], y0[k], x[lk], y[lk], l)

    dK.append(np.vstack([
        np.hstack([k11, k12]),
        np.hstack([k21, k22])
    ]))

    return dK


def build_dKdy0(x, y, x0, y0,l):
    # set up covariance matrix
    N = len(x)
    N0 = len(x0)
    k11 = np.empty((N0, N))
    k12 = np.empty((N0, N))
    k21 = np.empty((N0, N))
    k22 = np.empty((N0, N))

    for k in range(N0):
        for lk in range(N):
             k11[k,lk] = d3kdxdx0dy0(
                 x0[k], y0[k], x[lk], y[lk], l)
             k21[k,lk] = d3kdxdy0dy0(
                  x0[k], y0[k], x[lk], y[lk], l )
             k12[k,lk] = d3kdydx0dy0(
                  x0[k], y0[k], x[lk], y[lk], l)
             k22[k,lk] = d3kdydy0dy0(
                 x0[k], y0[k], x[lk], y[lk], l)

    K = np.vstack([
        np.hstack([k11, k12]),
        np.hstack([k21, k22])
    ])
    return K


def nlp_with_grad(hyp, xtrain, ytrain, ztrain, sig_n):
    K = np.empty((2*len(xtrain), 2*len(xtrain)))
    buildK(xtrain, ytrain, xtrain, ytrain, hyp, K)
    Ky = K + sig_n**2*np.diag(np.ones(np.shape(K)[0]))
    Kyinv = np.linalg.inv(Ky)                # invert GP matrix
    alpha = Kyinv.dot(ztrain)
    nlp_val = 0.5*ztrain.T.dot(alpha) + 0.5*np.linalg.slogdet(Ky)[1]

    dK = build_dK(xtrain, ytrain, xtrain, ytrain, hyp)

    nlp_grad = np.array([
        -0.5*alpha.T.dot(dK[0].dot(alpha)) + 0.5*np.trace(Kyinv.dot(dK[0])),
        -0.5*alpha.T.dot(dK[1].dot(alpha)) + 0.5*np.trace(Kyinv.dot(dK[1]))
    ])

    return nlp_val, nlp_grad

def nlp(hyp, xtrain, ytrain, ztrain, sig_n):
    #try:
    #    return nlpgen(hyp, xtrain, ytrain, ztrain, sig_n)
    #except:
    #    pass
    K = np.empty((2*len(xtrain), 2*len(xtrain)))
    buildK(xtrain, ytrain, xtrain, ytrain, hyp, K)
    Ky = K + sig_n**2*np.diag(np.ones(np.shape(K)[0]))
    Kyinv = np.linalg.pinv(Ky)                # invert GP matrix
    return 0.5*ztrain.T.dot(Kyinv.dot(ztrain)) + 0.5*np.linalg.slogdet(Ky)[1]

def gpsolve(Ky, ft):
    L = np.linalg.cholesky(Ky)
    alpha = solve_triangular(
        L.T, solve_triangular(L, ft, lower=True, check_finite=False),
        lower=False, check_finite=False)

    return L, alpha

def nlpgen(hyp, xtrain, ytrain, ztrain, sig_n):
    K = np.empty((2*len(xtrain), 2*len(xtrain)))
    buildK(xtrain, ytrain, xtrain, ytrain, hyp, K)
    # print(np.linalg.eigvals(K))
    Ky = K + sig_n**2*np.diag(np.ones(2*len(xtrain)))  # add noise variance
    L, alpha = gpsolve(Ky, ztrain)
    return 0.5*ztrain.T.dot(alpha) + np.sum(np.log(L.diagonal()))

def energy(x):
    return x[1]**2/2 + U0*(1 + np.cos(x[0]))

def calc_tau_ex(q,p):
    kappa = energy([q, p])/(2*U0)
    kappinv = 1/kappa
    if kappa < 1:
        tau = 4*scipy.special.ellipk(kappa)
    else:
        tau = 2.0*scipy.special.ellipk(kappinv)/np.sqrt(kappa)
    return tau

def buildKreg(xtrainreg, ytrainreg, lp):

    N = len(xtrainreg)
        # ztrain = np.concatenate((ztrain1, ztrain2))
    Kp = np.empty((N, N))

    for k in range(N):
        for lk in range(N):
            Kp[k,lk] = f_kern(
                     xtrainreg[k], ytrainreg[k], xtrainreg[lk], ytrainreg[lk], lp)
    return Kp

def build_dKreg(xtrainreg, ytrainreg, lp):

    N = len(xtrainreg)
        # ztrain = np.concatenate((ztrain1, ztrain2))
    Kp = np.empty((N, N))

    dK = []
    for k in range(N):
        for lk in range(N):
            Kp[k,lk] = dkdlx(
                     xtrainreg[k], ytrainreg[k], xtrainreg[lk], ytrainreg[lk], lp)

    dK.append(Kp.copy())

    for k in range(N):
        for lk in range(N):
            Kp[k,lk] = dkdly(
                     xtrainreg[k], ytrainreg[k], xtrainreg[lk], ytrainreg[lk], lp)

    dK.append(Kp.copy())

    return dK

def nlpreg(hyp, xtrain, ytrain, ztrain, sig_n):
    K = buildKreg(xtrain, ytrain, hyp)
    Ky = K + sig_n**2*np.diag(np.ones(np.shape(K)[0]))
    Kyinv = np.linalg.pinv(Ky)                # invert GP matrix
    return 0.5*ztrain.T.dot(Kyinv.dot(ztrain)) + 0.5*np.linalg.slogdet(Ky)[1]

def nlpreg_with_grad(hyp, xtrain, ytrain, ztrain, sig_n):
    K = buildKreg(xtrain, ytrain, hyp)
    dK = build_dKreg(xtrain, ytrain, hyp)
    Ky = K + sig_n**2*np.diag(np.ones(np.shape(K)[0]))
    Kyinv = np.linalg.inv(Ky)                # invert GP matrix
    alpha = Kyinv.dot(ztrain)

    nlp_val = 0.5*ztrain.T.dot(alpha) + 0.5*np.linalg.slogdet(Ky)[1]

    nlp_grad = np.array([
        -0.5*alpha.T.dot(dK[0].dot(alpha)) + 0.5*np.trace(Kyinv.dot(dK[0])),
        -0.5*alpha.T.dot(dK[1].dot(alpha)) + 0.5*np.trace(Kyinv.dot(dK[1]))
    ])

    return nlp_val, nlp_grad

def guessP(x, y, l, xtrainp, ytrainp, ztrainp, Kyinvp):
    Ntest = 1
    N = len(xtrainp)
    Kstar = np.empty((Ntest, N))
    for k in range(Ntest):
        for lk in range(N):
            Kstar[k,lk] = f_kern(
                 x[k], y[k], xtrainp[lk], ytrainp[lk], l)
    Ef = Kstar.dot(Kyinvp.dot(ztrainp))
    return Ef

def calcQ(x,y, xtrain, ytrain, l, Kyinv, ztrain):
    # get \Delta q from GP on mixed grid.
    # temporarily, also \delta p is passed from the function
    Kstar = np.empty((2, 2*len(xtrain)))
    buildK(xtrain, ytrain, [x], [y], l, Kstar)
    qGP = Kstar.dot(Kyinv.dot(ztrain))
    f = qGP[1]
    return f, qGP[0]

def Pnewton(P, x, y, l, xtrain, ytrain, Kyinv, ztrain):
    Kstar = np.empty((2, 2*len(xtrain)))
    buildK(xtrain, ytrain, x, P, l, Kstar)
    pGP = Kstar.dot(Kyinv.dot(ztrain))
    f = pGP[0] - y + P
    return f

def calcP(x,y, l, lp, xtrainp, ytrainp, ztrainp, Kyinvp, xtrain, ytrain, ztrain, Kyinv):
    # as P is given in an implicit relation, use newton to solve for P
    # use the second GP on regular grid (q,p) for a first guess for P
    pgss = guessP([x], [y], lp, xtrainp, ytrainp, ztrainp, Kyinvp)
    res, r = newton(Pnewton, pgss, full_output=True, maxiter=5, disp=False,
       args = (np.array([x]), np.array([y]), l, xtrain, ytrain, Kyinv, ztrain))
    return res

def applymap(l, lp, Q0map, P0map, xtrainp, ytrainp, ztrainp, Kyinvp, xtrain, ytrain, ztrain, Kyinv):
    Nq = Q0map.shape[0]
    Np = P0map.shape[1]
    #init
    pmap = np.zeros([nm, Nq, Np])
    pmaptest = np.zeros([nm, Nq, Np])
    qmap = np.zeros([nm, Nq, Np])
    H = np.zeros([nm, Nq, Np])
    #set initial conditions
    pmap[0,:,:] = P0map
    pmaptest[0, :, :] = P0map
    qmap[0,:,:] = Q0map
    H[0,:,:] = energy([Q0map, P0map])

    # loop through all points on grid (q, p) and all time steps
    for i in range(0,nm-1):
        for k in range(0, Nq):
            for lk in range(0, Np):
                # first: set new P
                pmap[i+1, k, lk] = calcP(qmap[i,k,lk], pmap[i, k, lk], l, lp, xtrainp, ytrainp, ztrainp, Kyinvp, xtrain, ytrain, ztrain, Kyinv)
                # pmap[1, k, lk] = pe0[k, lk]
                # pmap[2, k, lk] = p1[k, lk]
        for k in range(0, Nq):
            for lk in range(0, Np):
                if np.isnan(pmap[i+1, k, lk]):
                    qmap[i+1,k,lk] = np.nan
                else:
                    # then: set new Q via calculating \Delta q and adding q
                    qmap[i+1, k, lk], dpmap = calcQ(qmap[i,k,lk], pmap[i+1,k,lk],xtrain, ytrain, l, Kyinv, ztrain)# + qmap[i, k, lk]
                    qmap[i+1, k, lk] = np.mod(qmap[i+1,k,lk]+ qmap[i, k, lk], 2.0*np.pi)
                    # testing GP on mixed grid for p -> THIS does NOT work
                    # P = p + \delta p -> as the GP is fitted for -\delta p -> additional minus
                    pmaptest[i+1, k, lk] = pmap[i,k,lk] - dpmap
                H[i+1, k, lk] = energy([qmap[i+1, k, lk], pmap[i+1, k, lk]])
    return qmap, pmap, H, pmaptest

def dydt(y_, t):
    ydot = np.zeros([2])
    ydot[0] = y_[1] # dx/dt
    ydot[1] = -np.sin(y_[0] + np.pi)# - eps*np.sin(y_[0] - om*(t))#- 0.5*np.pi) # dpx/dt
    return ydot

def dydt_ivp(t, y):
    ydot = np.zeros([2])
    ydot[0] = y[1] # dx/dt
    ydot[1] = -np.sin(y[0]+np.pi)# - eps*np.sin(y_[0] - om*(t))#- 0.5*np.pi) # dpx/dt
    return ydot

def integrate_pendulum(q0, p0, t):
    # yint = np.zeros([len(t), 2, len(q0)]) # initial values for y
    ysint = np.zeros([len(t), 2, len(q0)]) # initial values for y

    # for k in range(len(q0)):
    #     yint[:, :, k] = spint.odeint(dydt, [q0[k], p0[k]], t)
        # ysint[:,:, k] = intode([q0[k], p0[k]], t, dtsymp)
        # ysint[:,0, k] = np.mod(ysint[:,0, k], 2*np.pi)
    ysint = []
    for ik in range(len(q0)):
        res_int = solve_ivp(dydt_ivp, [t[0], t[-1]], np.array((q0[ik], p0[ik])), max_step=0.001, method='RK45')
        temp = res_int.y
        # temp[0] = np.mod(temp[0], 2*np.pi)
        ysint.append(temp)

    return ysint
# compute log-likelihood according to RW, p.19
def solve_cholesky(L, b):
    return solve_triangular(
        L.T, solve_triangular(L, b, lower=True, check_finite=False),
        lower=False, check_finite=False)

# negative log-posterior
def nll_chol(hyp, x, y, build_K=build_K):
    K = np.empty((len(x), len(x)))
    build_K(x, x, hyp[:-1], K)
    Ky = K + np.abs(hyp[-1])*np.diag(np.ones(len(x)))
    L = np.linalg.cholesky(Ky)
    alpha = solve_cholesky(L, y)
    ret = 0.5*y.T.dot(alpha) + np.sum(np.log(L.diagonal()))
    #print(hyp, ret)
    return ret

def nll(hyp, x, y, neig=8, build_K=build_K):
    K = np.empty((len(x), len(x)))
    build_K(x, x, np.abs(hyp[:-1]), K)
    Ky = K + np.abs(hyp[-1])*np.diag(np.ones(len(x)))
    w, Q = eigsh(Ky, neig, tol=max(1e-6*np.abs(hyp[-1]), 1e-15))
    while np.abs(w[0]-hyp[-1])/hyp[-1] > 1e-6 and neig < len(x):
        if neig > 0.05*len(x):  # TODO: get more stringent criterion
            try:
                return nll_chol(hyp, x, y, build_K)

            except:
                print('Warning! Fallback to eig solver!')
        neig =  2*neig
        w, Q = eigsh(Ky, neig, tol=max(1e-6*hyp[-1], 1e-15))

    alpha = Q.dot(np.diag(1.0/w).dot(Q.T.dot(y)))

    ret = 0.5*y.T.dot(alpha) + 0.5*(np.sum(np.log(w)) + (len(x)-neig)*np.log(np.abs(hyp[-1])))
    #print(hyp, ret)
    return ret

def plot_pendulum(ysint):
    Np = ysint.shape[3]
    Nq = ysint.shape[2]
    plt.figure()
    for kp in range(Np):
        for kq in range(Nq):
            plt.plot(ysint[:,0,kq,kp], ysint[:,1,kq,kp],'.', color = 'k', markersize = 1.0)
            plt.plot(ysint[0,0,kq,kp],ysint[0,1,kq,kp],'o', color = 'b', markersize = 4.9)
            plt.plot(ysint[-1,0,kq,kp],ysint[-1,1,kq,kp],'o', color = 'r', markersize = 4.9)
    plt.xlabel(r'$q(n{\tau})$')
    plt.ylabel(r'$p(n{\tau})$')

def plot_map(qmap, pmap, H, Q0map, P0map):
    plt.figure()
    for i in range(0, qmap.shape[2]):
        plt.plot(qmap[:,:,i], pmap[:,:,i], ',', markersize = 1.5)
        # plt.plot(np.mod(qmap[:,:,i]-np.pi, 2*np.pi)-np.pi, pmap[:,:,i], ',', color = 'k', markersize = 1.5)
        # plt.plot(qmap_RBF[:,:,i], pmap_RBF[:,:,i], '.', color = 'k', markersize = 1.5)
    # plt.ylim([-4, 4])
    plt.xlabel('q')
    plt.ylabel('p')

    plt.figure()
    for i in range(0, qmap.shape[2]):
        # i = 0
        plt.plot(H[:,:,i]/energy([Q0map[:,i], P0map[:,i]]))#/H_RBF[0,:,i])
    plt.xlabel(r'$\tau$')
    plt.ylabel('E/E_0')

def gamma(d):
    x=1.0000
    for i in range(20):
        x = x-(pow(x,d+1)-x-1)/((d+1)*pow(x,d)-1)
    return x

def bluenoise(d,n, method, qmin, qmax, pmin, pmax):

    g = gamma(d)
    alpha = np.zeros(d)
    for j in range(d):
        alpha[j] = pow(1/g,j+1) %1
    z = np.zeros((n, d))
    # z = []
    i = 0
    iz = 0
    while iz < n:
        random_point = (0.5 + alpha*(i+1)) %1
        random_point[0] = random_point[0]*(np.abs(qmin)+qmax)-np.abs(qmin)
        random_point[1] = random_point[1]*(np.abs(pmin)+pmax)+pmin
        if (method == 'trapped') & (energy(random_point)/2 < 1.0):
            z[iz] = random_point
            iz = iz + 1
        elif (method == 'passing') & (energy(random_point)/2 > 1.0):
            z[iz] = random_point
            iz = iz + 1
        elif (method == 'all'):
            z[iz] = random_point
            iz = iz + 1
        i = i + 1
        #     while len(points) < num_points:
        # random_point = Point(temp[i,:])
        # if (random_point.within(poly)):
        #     points.append(random_point)
        # i = i + 1
    return z
